#
# Provides a BPMN to asynchronous pi-calculus parser
# Copyright (C) 2007 Frank Puhlmann.
#

require 'rexml/document' 
include REXML
require 'strscan'
require 'pimodel.rb'
require 'bpmnanalyzer.rb'

class BPMN2PiParser

	attr_reader :processes, :file, :agentList, :analyzers, :xml, :options

	def initialize(file,options=Hash.new)
		@file = file
		@agentList = Hash.new
		@processes = Array.new
		@analyzers = Array.new
		@options = options	
		parse(file)
	end
	
	# Parses a BPMN-XML file to pi-calculus
	def parse(file)
	
		puts "TODO: Implement complete structural soundness check!"
		
		@xml = Document.new(File.open(file))
		
		xml = @xml
				
		if xml.root.name != "model" 
    		raise "Cannot convert file, root element is not <model>!"
	  	end
	  	
	  	# Iterate through all processes and process all of type BPMN
		xml.elements.each("model/process[@type='BPMN']") do | process |
		
			process_id = process.attributes['id']
		
			puts "Processing BPD with id " + process_id + " (" + process.attributes['description'].to_s + ")..." 
			
			dag = detectCycles(xml,process_id)
			
			# Create global restrictions and components
			components = Array.new
			restrictions = Array.new
			
			# Create an agent for each node
			xml.elements.each("model/process[@id=#{process_id}]/node") do |n|
			
				#puts "Working on " + n.to_s
				
				identifiers = Array.new
				
			  	# Select all flow nodes that have this n.id as target
			  	flow_to = Array.new
			  	xml.elements.each("/model/process[@id=#{process_id}]/flow[@to='#{n.attributes['id']}'][@type='Sequence Flow']") do |s| 
			  		flow_to << s
			  		identifiers << "e" + s.attributes['id']
			  		end
			  
				# Select all flow nodes that have this n.id as source
			  	flow_from = Array.new
			  	xml.elements.each("/model/process[@id=#{process_id}]/flow[@from='#{n.attributes['id']}'][@type='Sequence Flow']") do |s| 
			  		flow_from << s
			  		identifiers << "e" + s.attributes['id']
			  	end
			  	
			  	#puts "In = " + flow_to.inspect + " Out = " + flow_from.inspect
			  	
			  	# Create local variable for storing parameter values
			  	params = Array.new
			  	# Create a new agent
			  	agent_id = n.attributes['id']
			  	nextAgent = Agent.new("N" + process_id + "_" + agent_id,nil,nil)	
			  	nextAgent.description = n.attributes['type']	 
			 	nextNode = nextAgent			
			 	# Additional agent
			 	extraAgent = nil		  	
			  	
			  	#
			  	# Apply pattern depending on node type
			  	#
			  	case n.attributes['type']
			  	
			  	when 'Start Event', 'End Event', 'Task', 'Intermediate Event':
			  	
			  		#puts "Processing " + n.attributes['type'] + "..."
			  		
			  		# Check if max. one in- and one outgoing sequence flow
			  		if flow_from.size > 1 
			  			raise "PARSE EXCEPTION: " + n.to_s + " has more than one incoming sequence flow!"
			  			end
			  		if flow_to.size > 1 
			  			raise "PARSE EXCEPTION: " + n.to_s + " has more than one incoming sequence flow!"
			  			end			  		
			  	
			  		# Skip if Intermediate event has 'scope' attribute
			  		if n.attributes['type'] == 'Intermediate Event'
			  			next if n.attributes['scope']
			  			end

				  	# Check if TASK with attached SCOPE EVENT
				  	se_list = Array.new
				  	se_list = xml.elements.each("/model/process[@id=#{process_id}]/node[@scope='#{n.attributes['id']}']") do |se| 
				  		se_list << se
				  		end
	  					  					  					  	
			  		if flow_to.size == 1
			  			nextNode = appendNode(nextNode, Input.new("e" + flow_to[0].attributes['id']))
			  			params << nextNode.object
			  			end	
			  			
				  	if se_list.size > 0  
				  		# Scope Events attached!
				  		comp = Composition.new()			
				  		# Add <.>.'check.0 part	  		
				  		fabstr = fabstr(nextAgent.ident)
				  		fabstr.successor = Output.new('check')
				  		comp.parts << fabstr
				  		# Add parts for each attached event to summation
				  		sum = Summation.new
				  		comp.parts << sum
				  		se_list.each do |p|
				  			event_to = xml.elements.to_a("/model/process[@id=#{process_id}]/flow[@from='#{p.attributes['id']}']")[0].attributes['id']
				  			fabstr2 = fabstr(p.attributes['id'].to_s)	
				  			fabstr2.successor = Output.new('e' + event_to)
				  			# Check for cycles
				  			if !dag && flow_to.size > 0
					  			fabstr2.successor = Composition.new(AgentRef.new(nextAgent,nextAgent.params),
					  				fabstr2.successor)
					  			end				  					  			
				  			sum.parts << fabstr2
				  			params << 'e' + event_to
				  			end
				  		# Add default behavior to sum				  		
				  		default = Input.new('check',Output.new("e#{flow_from[0].attributes['id']}"))
				  		# Check for cycles
				  		if !dag && flow_to.size > 0
					  		default.successor = Composition.new(AgentRef.new(nextAgent,nextAgent.params),
					  			default.successor)
					  		end
				  		
				  		sum.parts << default 
				  		params << "e#{flow_from[0].attributes['id']}"
				  		
				  		nextNode = appendNode(nextNode,Restriction.new('check'))
				  		nextNode = appendNode(nextNode,comp)
				  		
				  		else			  			
			  			
				  		nextNode = appendNode(nextNode,fabstr(nextAgent.ident))	
						nextNode = insertCycle(nextNode,nextAgent) if !dag && flow_to.size > 0
			  					  		
				  		if flow_from.size == 1
				  			nextNode = appendNode(nextNode, Output.new('e' + flow_from[0].attributes['id']))
			  				params << nextNode.object
			  				end
			  			end
	
			  				
			  	when 'AND Gateway':
			  	
					#puts "Processing " + n.attributes['type'] + "..."								  		
					
					flow_to.size.times do |f|
						nextNode = appendNode(nextNode, Input.new("e" + flow_to[f].attributes['id']))
						params << nextNode.object
					end
			  		nextNode = appendNode(nextNode,fabstr(nextAgent.ident))		
			  		comp = Composition.new	  
			  		comp.parts << AgentRef.new(nextAgent,nextAgent.params) if !dag && flow_to.size > 0		
			  		flow_from.size.times do |f|
			  			out = Output.new("e" + flow_from[f].attributes['id'])
			  			comp.parts << out
			  			params << out.object
			  		end		
			  		nextNode = appendNode(nextNode,comp)
			  		
			  	when 'Gateway', 'XOR Gateway':
			  	
					#puts "Processing " + n.attributes['type'] + "..."								  		
					
					# Create agent for outgoing flows
					agentCont = AgentRef.new(Agent.new(nextAgent.ident + "_1",nil,nil),nil)
					agentCont.agent.description = n.attributes['type'] + " (second part)"
					paramsCont = Array.new
					nextNode = agentCont
			  		#nextNode = appendNode(nextNode,Tau.new)
			  		sum = Summation.new	  
			  		flow_from.size.times do |f|
			  			tau = fabstr(nextAgent.ident)
			  			out = Output.new("e" + flow_from[f].attributes['id'])			  			
			  			if !dag && flow_to.size > 0
			  				tau.successor = Composition.new(AgentRef.new(nextAgent,nextAgent.params),out)
			  				else
			  				tau.successor = out
			  				end
			  			sum.parts << tau
			  			paramsCont << out.object
			  			params << out.object
			  		end		
			  		nextNode = appendNode(nextNode,sum)		
			  		agentCont.agent.params = paramsCont
			  							
					# Create summation with each incoming flow
					sum = Summation.new
					flow_to.size.times do |f|
						nextNode = appendNode(sum, Input.new("e" + flow_to[f].attributes['id']))
						params << nextNode.object
						# add objects to outgoing flows agent
						agentCont.agent.params << nextNode.object
						appendNode(nextNode,agentCont)
					end
					nextAgent.behavior = sum
					
			  		extraAgent = agentCont.agent
			  		
			  	when 'OR Gateway':
			  	
			  		puts "Processing " + n.attributes['type'] + "..."
			  		
			  		# Check if OR-Join required
			  		if flow_to.size > 1

#			  			nextNode = appendNode(nextNode, Restriction.new("or_done"))
#			  			nextNode = appendNode(nextNode, Composition.new)
#						orComp = Composition.new			  			
#						flow_to.size.times do |f|
#							orComp.parts << Input.new("e" + flow_to[f].attributes['id'])
#							params << orComp.parts[-1].object
#							orComp.parts[-1].successor =Output.new('or_done')
#							end
#						appendNode(nextNode, orComp)		
#						inp = Input.new('or_done')
#						inp.function=(nextAgent.ident)
#						nextNode = appendNode(nextNode, inp)

						nextNode = appendNode(nextNode, Restriction.new("or_done"))
						nextNode = appendNode(nextNode, Restriction.new("cancel"))
						nextNode = appendNode(nextNode, Restriction.new("ack"))
						orComp = Composition.new
						nextNode = appendNode(nextNode, orComp)
						# Create left hand part
						leftPart = Composition.new
						flow_to.size.times do |f|
							inp = Input.new('cancel',Output.new('ack'))
							inp.function=(nextAgent.ident)
							sum = Summation.new(Input.new("e" + flow_to[f].attributes['id'],
									Composition.new(Output.new('or_done'),Output.new('ack'))),
									inp)
							leftPart.parts << sum	
							params << "e" + flow_to[f].attributes['id']
							end					
						orComp.parts << leftPart	
						# Create right hand part
						rightPart = Composition.new
						(flow_to.size - 1).times do |f|
							rightPart.parts << Output.new('cancel')
							end
						seqStart = Input.new('ack')
						seqStart.function=(nextAgent.ident) 
						nextNode = seqStart
						(flow_to.size - 1 ).times do |f|
							inp = Input.new('ack')
							inp.function=(nextAgent.ident)
							nextNode = appendNode(nextNode,inp)
							end
						rightPart.parts << seqStart
						inp = Input.new('or_done', rightPart)
						inp.function=(nextAgent.ident)
						orComp.parts << inp
									
						else
							raise "OR JOIN with less than two arcs not yet supported!"
						end
			  		nextNode = appendNode(nextNode,fabstr(nextAgent.ident))		
			  		
			  		# Handle Multiple Choice
			  		if flow_from.size > 1
			  				raise "MULTI CHOICE not yet supported"
			  			else
						nextNode = insertCycle(nextNode,nextAgent) if !dag 
						nextNode = appendNode(nextNode, Output.new('e' + flow_from[0].attributes['id']))
			  			params << nextNode.object
			  			end
			  	
			  	else 
			  		puts "Unkown node type ignored: " + n.to_s 
			  		
			    end
			    	  	
			  	nextAgent.params=params.flatten
			  	restrictions << params
			  	components << AgentRef.new(nextAgent,nextAgent.params)
			  		
			  	puts nextAgent.to_definition
				
				puts extraAgent.to_definition if extraAgent
			
			end
			
			#puts restrictions.flatten.uniq.inspect
			
			# Create global agent
			agent = AgentRef.new(Agent.new("N" + process_id, nil, nil),nil)
			# Add description
			agent.agent.description = process.attributes['description'].to_s
			# Chain restrictions
			nextNode = agent
			restrictions.flatten.uniq.each do |r|
				nextNode = appendNode(nextNode,Restriction.new(r,nil))
			end
			# Add components
			componentNode = Composition.new()
			componentNode.parts = components
			appendNode(nextNode,componentNode)
			# Add to agent list
			agentList["N" + process_id] = agent			
			
			puts agent.agent.to_definition
			puts "--------------"
			
			# Set default agent to last global agent
			@processes << agent
		
		end		
		
		return processes
	end
	
private

	def fabstr(function)
		tau = Tau.new
		tau.function = function
		return tau
	end

	def detectCycles(xml,process_id)
	
		@analyzers << BPMNAnalyzer.new(xml,process_id)
		tc = @analyzers[-1].tcMatrix
		#@analyzers[-1].printMatrix(tc)
		
		# if (1) one initial node, 
		
		# (2) one final node, 
		
		# (3) each node is on a path from i to o
		
		# (4) cycles are contained
		dagFlag = true
		tc.size.times do |i|
			dagFlag = false if tc[i][i] < 9999
			end
			
		puts "DAG: " + dagFlag.to_s
		
		if @options['forceCycles'] == "true"
			
			puts "ENFORCING CYCLES"
			
			dagFlag = false
			
			end
	
		return dagFlag
	
	end
	

	
	def insertCycle(nextNode,nextAgent)
	
		# puts "INSERT CYCLE"
	
		# Handle cycles
		c = Composition.new(AgentRef.new(nextAgent,nextAgent.params))
		nextNode = appendNode(nextNode,c)
		return nextNode
	end
	


	def appendNode(beh,node)
	
		if !beh
			beh = node
			return node
		end
		
		case beh 
		
			when Composition, Summation:
				beh.parts << node
				
			when Input, Output, Tau, Restriction:
				beh.successor = node	
				
			when Agent:
				beh.behavior = node
				
			when AgentRef:
				beh.agent.behavior = node

			else
				raise "Unkown node type to append!"
		end
		
		return node
	
	end
	

end