#
# Provides a pi-calculus data model
# Copyright (C) 2007 Frank Puhlmann.
#

class Agent 

    attr_accessor:ident, :behavior, :params, :restrictions, :description

    def initialize(ident, params, behavior)
        @ident = ident
        @behavior = behavior
        @params = params
        @restrictions = Array.new
    end
    
    def to_s
        return ident + "(" + params.join(",") + ")" if params != nil && params.size>0
        ident
    end
    
    def to_definition
    	res = ""
    	res = restrictions.join(",") if restrictions
    	if (res.size > 0) 
    		res = "(^" + res + ")"
    		end if restrictions
        return to_s + "=" + res + behavior.to_s if behavior
        return to_s + "=0"
    end
    
    # Returns the bound names of this agent
    def bn
        bound=Array.new
    	_bn(behavior,bound).flatten
    end
    
    def substitute(params2)
    
    	puts "SUBSTITUTING " + params.inspect + " with " + params2.inspect
    
        if params.size != params2.size 
            raise "Parameter size mismatch in substitution applied to " + to_s + "!"
        end
        
        bound=Array.new
    
        params2.size.times do |i| 
        	_subst(@behavior,bound,params[i],params2[i])  if params[i] != params2[i]
        	end
        
        @params = params2
        
    end
    
	#
	# Prints the execution tree of the agent
	#
	def to_tree
		_printTree(self)
	end
	
	def _printTree(agent,*l)
	
		if l.size == 0
			level = 0
			@visited = Array.new
			else
			level = l[0].to_i
			end
		
		level.times do 
			print " "
			end
		print "-" + agent.class.to_s
		print " " + agent.to_s if agent.class == Agent
		print " " + agent.to_s if agent.class == AgentRef
		print " " + agent.object.to_s + "(" + agent.subjects.join(",").to_s + ")" if agent.class == Input
		print " '" + agent.object.to_s + "<" + agent.subjects.join(",").to_s + ">" if agent.class == Output
		puts		
	
		case agent
		
			when Input, Output, Tau, Restriction:
				_printTree(agent.successor,level+1) if agent.successor
						
			when Summation, Composition:
				
				agent.parts.each do |p|
					_printTree(p,level+1) 
					end
					
			when AgentRef:

				if !@visited.include?(agent)
					@visited << agent
					_printTree(agent.agent.behavior,level+1) 
					end
					
			when Agent:
			
				if !@visited.include?(agent)
					@visited << agent
					_printTree(agent.behavior,level+1) 
					end
				
			end

	
	end
   
        
    # Calculates the bound names of beh, with an initial set of bound names given
    # (Ignores names bound in defined agents!)
    def _bn(beh,bound)
    	# Bound inputs have to be considered
    	case beh
    	
    	when Input:
    	    	
            if beh.subjects.size > 0 
                (bound << beh.subjects).flatten!.uniq!                                      
            end
            
        when Restriction:
                
        	bound << beh.name    
            
    	end
    	
    	# Continue recursivly
    	case beh
        
        when Composition, Summation:

            beh.parts.each { |p| _bn(p,bound)}
    	
    	when Output, Input, Tau, Restriction:
           _bn(beh.successor,bound) if beh.successor
        end
        
        bound
        
    end
    
    # Substitutes free names in an agent, respecting bound ones
    def _subst(beh,glob_bound,from,to)
    
    # puts "Subst: " + beh.to_s + " {"+to+"/"+from+"}"
    
      case beh
        
        when Composition, Summation:

			# Clone global_bound names before!
            beh.parts.each { |p| _subst(p,glob_bound.clone,from,to)}
            
        when Input, Output: 

			# Get all currently bound names
			bound = _bn(beh,glob_bound)
			
			#puts "BOUND: " + bound.inspect + ", " + glob_bound.inspect + " ( " + beh.to_s + " )"
            
            # Replace object only if not bound(!)
            if !bound.include?(beh.object)
                beh.object = to if beh.object == from 
            end
        
            beh.subjects.size.times do |i| 
                # Replace subject only if not bound(!)
                if !bound.include?(beh.subjects[i-1])
                    beh.subjects[i-1] = to if beh.subjects[i-1] == from                                  
                end
            end if beh.subjects 
                
            _subst(beh.successor,glob_bound,from,to) if beh.successor
            
        when AgentRef:
        	if beh.parameters
 	        	# Get all currently bound names
				#bound = _bn(beh,glob_bound)
    	       
	        	beh.parameters.size.times do |i|
	        		# Replace params only if not bound(!)
	        		#if !bound.include?(beh.parameters[i-1])
	        			beh.parameters[i-1]=to if beh.parameters[i-1] == from
	        		#end
	        	end  
        	
        	end
        
        when Agent:
	        # Get all currently bound names
			#bound = _bn(beh,glob_bound)
    
            beh.params.size.times do |i| 
                # Replace params only if not bound(!)
                #if !bound.include?(beh.params[i-1])
                    beh.params[i-1]=to if beh.params[i-1] == from 
                #end
            end
     
        when Tau, Restriction:
	        # move on
	        _subst(beh.successor,glob_bound,from,to) if beh.successor
        end 
        
    end 
    
end 

class AgentRef

	attr_accessor :agent
	
	def initialize(agent,parameters)
		@agent = agent
		@parameters = parameters
	end
	
	def to_s
		params = ""
		params = parameters.join(",") if parameters
		params = "(" + params + ")" if params.size > 0
		agent.ident + params
	end
	
	def parameters
		@parameters = agent.params.clone if @parameters == nil && agent.params != nil
		return @parameters
	end
	
	def parameters=(params)
		@parameters = agent.params.clone if @parameters == nil && agent.params != nil
		@parameters = params
	end

end

class Restriction

	attr_accessor :name, :successor
	
	def initialize(name,successor=nil)
		@name = name
		@successor = successor
	end
	
	def to_s
	    succ = "#{successor ? successor : '0'}"
		return "(^" + name + ")" + succ
	end

end

class Composition

    attr_accessor :parts
    
    def initialize(*parts)
        @parts = parts
    end

    def to_s
        return "(" + parts.each{ |c| c.to_s }.join(" | ") + ")" if parts && parts.size>1
        return parts[0].to_s if parts && parts.size==1
        "0"
    end
    
end 

class Summation

    attr_accessor :parts
    
    def initialize(*parts)
        @parts = parts
    end

    def to_s
        return "(" + parts.each{ |c| c.to_s }.join(" + ") + ")" if parts && parts.size>1
        return parts[0].to_s if parts && parts.size==1
        "0"
    end

end

class Output

    attr_accessor :object, :subjects, :successor, :function
        
    def initialize(object, *subjects)
        @object = object
        @subjects = subjects
        subjects.flatten! if subjects
        if subjects.size > 0
            if !subjects[-1].instance_of?(String) 
            	puts "HERE " + subjects[-1].class.to_s
                @successor = subjects[-1] 
                subjects.slice!(-1)
            end
        end
        subjects.flatten! if subjects
    end
    
    def clone
    	out = Output.new(@object.clone)
    	out.subjects = @subjects.clone if @subjects
    	out.function = @function.clone if @function
    	return out    
    end
    
    def to_s
        subjects.flatten!       
        succ = ".#{successor ? successor : '0'}"
        return "'" + object + "<" + subjects.join(",") + ">" + succ if subjects.size > 0
        "'" + object + succ
    end

end

class Input 

    attr_accessor :object, :subjects, :successor, :function
    
    def initialize(object, *subjects)
        @object = object
        @subjects = subjects
        subjects.flatten! if subjects
        if subjects.size > 0
            if !subjects[-1].instance_of?(String) 
                @successor = subjects[-1] 
                subjects.slice!(-1)
            end
        end
    end
    
    def clone
    	inp = Input.new(@object.clone)
    	inp.subjects = @subjects.clone if @subjects
    	inp.function = @function.clone if @function
    	return inp
    end

    def to_s
        subjects.flatten!
        succ = ".#{successor ? successor : '0'}"
        return object + "(" + subjects.join(",") + ")" + succ if subjects.size > 0
        object + succ
    end

end

class Tau

    attr_accessor :successor, :function

    def initialize(*successor)
        @successor = successor[0] if successor.size > 0
    end
    
    def clone
    	tau = Tau.new(@sucessor)
    	tau.function = @function.clone if @function
    	return tau
    end
    
    def to_s
#    	if function
#    		"t[#{function.to_s}].#{successor ? successor : '0'}"	
#	    	else
	        "t.#{successor ? successor : '0'}"
#	        end
    end

end 




